from yta_general_utils.web.tiktok.url_parser import TiktokUrlParser
from yta_general_utils.dataclasses import FileReturn
from yta_general_utils.downloader.utils import get_file
from yta_general_utils.file.enums import FileTypeX
from yta_general_utils.programming.output import Output
from yta_general_utils.programming.validator import PythonValidator
from yta_general_utils.file.writer import FileWriter
from typing import Union
from io import BytesIO

import requests


def _get_tiktok_video_content(
    url: str
    # TODO: If possible, set the real return type
) -> any:
    """
    Get the Tiktok video with the given 'url' as a
    content buffer that can be written into a file
    (locally or in memory).

    For internal use only.

    This method has been found by scanning this
    platform:
    - https://tikcd.com/
    """
    if not PythonValidator.is_string(url):
        raise Exception('The provided "url" parameter is not a string.')
    
    tiktok_video_info = TiktokUrlParser.parse(url)

    # TODO: The old way is not working...
    headers = {
        'accept': '*/*',
        'accept-language': 'es-ES,es;q=0.9',
        'origin': 'https://tikcd.com',
        'priority': 'u=1, i',
        'referer': 'https://tikcd.com/',
        'sec-ch-ua': '"Chromium";v="134", "Not:A-Brand";v="24", "Google Chrome";v="134"',
        'sec-ch-ua-mobile': '?0',
        'sec-ch-ua-platform': '"Windows"',
        'sec-fetch-dest': 'empty',
        'sec-fetch-mode': 'cors',
        'sec-fetch-site': 'cross-site',
        'user-agent': 'Mozilla/5.0 (Windows NT 10.0; Win64; x64) AppleWebKit/537.36 (KHTML, like Gecko) Chrome/134.0.0.0 Safari/537.36',
    }

    # We force it to be large url
    params = {
        'url': tiktok_video_info.url,
        'hd': '1',
    }

    # TODO: This method has been found using this platform
    # https://tikcd.com/?srsltid=AfmBOopX4VP6_pfbANSdRaXUfnXXDDo6zIHlzgbl_QsfBR0tmWyhXjPh
    # It seems that the old tikcdn doesn't work

    return get_file(
        requests.get(
            'https://tikwm.com/api/',
            params = params,
            headers = headers
            # TODO: I should handle this if unavailable
        ).json().get('data').get('wmplay')
    )

"""
Exposed methods below.
"""

def download_tiktok_video(
    url: str,
    output_filename: Union[str, None, bool] = None
) -> Union[FileReturn, BytesIO]:
    """
    Get the Tiktok video with the given 'url' as a file
    that is downloaded and stored locally with the also
    provided 'output_filename' (or a temporary one if
    not provided). If 'output_filename' is provided, the
    file is returned as a file in memory (BytesIO).
    """
    if not PythonValidator.is_string(url):
        raise Exception('The provided "url" parameter is not a string.')
    
    file_in_memory = BytesIO(_get_tiktok_video_content(url))

    return (
        file_in_memory
        if output_filename is False else
        FileWriter.write_binary_file(
            file_in_memory.getvalue(),
            Output.get_filename(output_filename, FileTypeX.VIDEO)
        )
    )


# TODO: Make this work also with video_id (?)
# TODO: This is no longer working, I think. Maybe
# restablished in a near future...
# def get_tiktok_video_old(
#     url: str,
#     output_filename: Union[str, None] = None
# ) -> FileReturn:
#     """
#     Obtains the Tiktok video from the provided 'url' if 
#     valid and stores it locally if 'output_filename' is
#     provided, or returns it if not.
#     """
#     DOWNLOAD_CDN_URL = 'https://tikcdn.io/ssstik/' # + video_id to download
#
#     if not PythonValidator.is_string(url):
#         raise Exception('The provided "url" parameter is not a string.')

#     tiktok_video_info = TiktokUrlParser.parse(url)

#     download_url = f'{DOWNLOAD_CDN_URL}{tiktok_video_info.video_id}'

#     return download_video(
#         download_url,
#         Output.get_filename(output_filename, FileTypeX.VIDEO)
#     )


"""
This file has been commented to delete the code in a near future
and avoid the use of 'cv2' so we can remove it as a requirement.
"""

# import cv2


# def detect_faces(image_filename):
#     """
#     This method receives the 'image_filename' image input and detects the faces that
#     are in that image. It returns an array with all existing faces, with their coordinates
#     and the width and height of the rectangle that surrounds the face.
#     """
#     # I used this: https://www.datacamp.com/tutorial/face-detection-python-opencv
#     img = cv2.imread(image_filename)
#     gray_image = cv2.cvtColor(img, cv2.COLOR_RGB2GRAY)

#     face_classifier = cv2.CascadeClassifier(
#         cv2.data.haarcascades + 'haarcascade_frontalface_default.xml'
#     )

#     face = face_classifier.detectMultiScale(
#         gray_image, scaleFactor = 1.1, minNeighbors = 5, minSize = (40, 40)
#     )

#     faces_detected = []

#     for (x, y, w, h) in face:
#         faces_detected.append({
#             'x': x,
#             'y': y,
#             'width': w,
#             'height': h,
#         })
#         #cv2.rectangle(img, (x, y), (x + w, y + h), (0, 255, 0), 4)

#     #cv2.imwrite('wip/screenshot_with_faces.png', img)
        
#     return faces_detected
    
# def detect_humans(image_filename):
#     """
#     This method receives the 'image_filename' image input and detects the humans that
#     are in that image. It returns an array with all existing humans, with their coordinates
#     and the width and height of the rectangle that surrounds the human.

#     This works only for human full bodies.
#     """
#     # Reading the Image
#     image = cv2.imread(image_filename)

#     # initialize the HOG descriptor
#     hog = cv2.HOGDescriptor()
#     hog.setSVMDetector(cv2.HOGDescriptor_getDefaultPeopleDetector())

#     # detect humans in input image
#     (humans, _) = hog.detectMultiScale(image, winStride = (10, 10), padding = (32, 32), scale = 1.1)

#     # getting no. of human detected
#     print('Human Detected : ', len(humans))

#     humans_detected = []

#     # loop over all detected humans
#     for (x, y, w, h) in humans:
#         humans_detected.append({
#             'x': x,
#             'y': y,
#             'width': w,
#             'height': h,
#         })
#         pad_w, pad_h = int(0.15 * w), int(0.01 * h)
#         cv2.rectangle(image, (x + pad_w, y + pad_h), (x + w - pad_w, y + h - pad_h), (0, 255, 0), 2)

#     cv2.imwrite('wip/test_human_detection.png', image)

#     return humans_detected


from yta_general_utils.programming.validator import PythonValidator
from yta_general_utils.programming.validator.number import NumberValidator
from yta_general_utils.math.value_normalizer import ValueNormalizer


class NormalizableValue:
    """
    Class to represent a value within a range, useful
    to normalize or denormalize it without doubt if the
    value is yet normalized or not.

    We store any value as not normalized but you are
    able to normalize it in any case as we know the 
    range.
    """

    value: float = None
    """
    The value not normalized.
    """
    range: tuple[float, float] = None
    """
    The range in which the 'value' is contained when
    not normalized so we can use it to normalized it.
    """

    @property
    def normalized(self) -> float:
        """
        The 'value' but normalized according to the 'range' in
        which the lower limit will by represented by the 0 and
        the upper limit by the 1.
        """
        return ValueNormalizer(self.range[0], self.range[1]).normalize(self.value)

    def __init__(
        self,
        value: float,
        range: tuple[float, float],
        value_is_normalized: bool = False
    ):
        """
        Initialize the 'value' within the 'range' provided. If
        the 'value' provided is already normalized, set the 
        'value_is_normalized' flag as True to be correctly
        recognized.
        """
        if not PythonValidator.is_tuple(range) or len(range) != 2:
            raise Exception('The provided "range" is not a tuple of 2 values.')
        
        if range[0] > range[1]:
            raise Exception('The provided "range" first value is greater or equal to the second one.')
        
        if not value_is_normalized and not NumberValidator.is_number_between(value, range[0], range[1]):
            raise Exception('The provided "value" is out of the given "range".')
        elif value_is_normalized and not NumberValidator.is_number_between(value, 0, 1):
            raise Exception('The provided "value" is out of the normalized range [0, 1].')
        
        # Denormalize the value if the one provided is normalized
        value = ValueNormalizer(range[0], range[1]).denormalize(value) if value_is_normalized else value
        
        self.value = value
        self.range = range


"""
This file has been commented to delete the code in a near future
and avoid the use of 'cv2' so we can remove it as a requirement.
"""

# from yta_general_utils.temp import Temp
# from PIL import Image

# import os


# # Interesting: http://www.leancrew.com/all-this/2013/11/transparency-with-pil/
# # Also this: https://www.101computing.net/pixel-art-in-python/
# # Try this to save drawing Pixel Arts: https://stackoverflow.com/questions/41319971/is-there-a-way-to-save-turtles-drawing-as-an-animated-gif
 
# PROJECT_ABSOLUTE_PATH = os.getenv('PROJECT_ABSOLUTE_PATH')
# TOUSE_ABSOLUTE_PATH = os.getenv('TOUSE_ABSOLUTE_PATH')
# FONTS_PATH = 'C:/USERS/DANIA/APPDATA/LOCAL/MICROSOFT/WINDOWS/FONTS/'

# def test_minecraft():
#     # This can turn a video into a pixel art video, amazing
#     #test_pixelart('C:/Users/dania/Downloads/nico.MOV', 'test_pixelart.mp4')

#     ICON_FILENAME = TOUSE_ABSOLUTE_PATH + 'minecraft_resources/fav.png'

#     icon = Image.open(ICON_FILENAME)
#     img = Image.new('RGB', (icon.width, icon.height))

#     for x in range(img.width):
#         for y in range(img.height):
#             img.putpixel((x, y), (icon.getpixel((x, y))))
#             img.save(Temp.create_filename('tmp_pixel_' + str(x) + '_' + str(y) + '.png'))

#     # Try this to preview (?) (https://stackoverflow.com/questions/42719095/how-to-show-an-image-with-pillow-and-update-it)
    
# def live_preview():
#     import numpy as np
#     import cv2

#     def sin2d(x,y):
#         """2-d sine function to plot"""
#         return np.sin(x) + np.cos(y)

#     def getFrame():
#         """Generate next frame of simulation as numpy array"""

#         # Create data on first call only
#         if getFrame.z is None:
#             xx, yy = np.meshgrid(np.linspace(0,2*np.pi,w), np.linspace(0,2*np.pi,h))
#             getFrame.z = sin2d(xx, yy)
#             getFrame.z = cv2.normalize(getFrame.z,None,alpha=0,beta=1,norm_type=cv2.NORM_MINMAX, dtype=cv2.CV_32F)

#         # Just roll data for subsequent calls
#         getFrame.z = np.roll(getFrame.z,(1,2),(0,1))
#         return getFrame.z

#     # Frame size
#     w, h = 640, 480

#     getFrame.z = None

#     while True:
#         # Get a numpy array to display from the simulation
#         npimage=getFrame()

#         cv2.imshow('image', npimage)
#         cv2.waitKey(1)




from yta_general_utils.downloader.utils import download_file
from yta_general_utils.programming.output import Output
from yta_general_utils.file.enums import FileTypeX
from yta_general_utils.dataclasses import FileReturn
from typing import Union


def download_video(
    url: str,
    output_filename: Union[str, None] = None
) -> FileReturn:
    """
    Receives a downloadable url as 'url' and downloads that video in
    our system as 'output_filename'.
    """
    return download_file(url, Output.get_filename(output_filename, FileTypeX.VIDEO))


from dataclasses import dataclass


@dataclass
class TiktokUrl:

    username: str
    """
    The user who the video belongs to.
    """
    video_id: str
    """
    The id of the video.
    """
    url: str
    """
    The long Tiktok url of that video.
    """

    def __init__(
        self,
        username: str,
        video_id: str,
        url: str
    ):
        self.username = username
        self.video_id = video_id
        self.url = url


class Node:
    """
    Class that represent a Node in a Graphic, which
    has to be inside the limits (this is checked by
    the Graphic when added to it).
    """
    
    position: tuple[float, float]

    @property
    def x(self):
        return self.position[0]

    @property
    def y(self):
        return self.position[1]

    def __init__(self, x: float, y: float):
        self.position = (x, y)


from dataclasses import dataclass


@dataclass
class SubtitleLine:
    """
    Class to represent a line of a subtitles file.
    """
    
    text: str
    """
    The text that is displayed in the subtitle line.
    """
    start_time: int
    """
    The time moment (in ms) in which the subtitle 
    line starts being displayed.
    """
    duration: int
    """
    The time (in ms) that the subtitle line last
    being displayed.
    """

    @property
    def end_time(self) -> int:
        """
        The time moment (in ms) in which the subtitle
        line ends being displayed.
        """
        return self.start_time + self.duration

    def __init__(
        self,
        text: str,
        start_time: int,
        duration: int
    ):
        self.text = text
        self.start_time = start_time
        self.duration = duration

@dataclass
class Subtitles:
    """
    Class to represent the subtitles of a file.
    """

    subtitles: list[SubtitleLine]
    """
    The list of subtitles line.
    """

    @property
    def text(
        self
    ) -> str:
        """
        Get the whole subtitles text by concatenating
        all the subtitle lines.
        """
        return '\n'.join([
            subtitle.text.strip()
            for subtitle in self.subtitles
        ])

    def __init__(
        self,
        subtitles: list[SubtitleLine] = []
    ):
        self.subtitles = subtitles

    def add_subtitle(
        self,
        subtitle: SubtitleLine
    ):
        self.subtitles.append(subtitle)


"""
This file has been commented to delete the code in a near future
and avoid the use of 'cv2' so we can remove it as a requirement.
"""

# from yta_general_utils.temp import create_tmp_filename
# from yta_general_utils.video_processor import extract_audio_from_video
# from moviepy import VideoFileClip, CompositeVideoClip, CompositeAudioClip, AudioFileClip
# from yta_general_utils.experimental.tiktok import download_tiktok_video, process_tiktok_video_url

# import numpy as np
# import os
# import random
# import cv2
# import math


# # TODO: This should be moved to multimedia because
# # it is video related
# def build_short_from_tiktok_video(tiktok_url, overlay_video_filename):
#     """
#     This method receives a tiktok video url and creates a new video to be shared as short
#     that contains the original video and an overlay gameplay or content to customize it.
#     """
#     OVERLAY_CROPPED_VIDEO = create_tmp_filename('overlay_cropped.mp4')
#     OVERLAY_RESCALED_VIDEO = create_tmp_filename('overlay_rescaled.mp4')
    
#     # Check if exist and has been processed to avoid processing it again
#     tiktok_video_data = process_tiktok_video_url(tiktok_url)
#     # database_tiktok_video = db_select('tiktok_videos', 'video_id', tiktok_video_data['video_id'])
#     # if database_tiktok_video and database_tiktok_video['status'] == 'generated':
#     #     print('The video ' + tiktok_video_data['video_id'] + ' from user @' + database_tiktok_video['username'] + ' has already been uploaded. Stopping process...')
#     #     return
    
#     output_video_filename = tiktok_video_data['username'] + '_' + tiktok_video_data['video_id'] + '_one.mp4'

#     # *. Download tiktok video
#     downloaded_video_filename = download_tiktok_video(tiktok_url)
#     main_video = VideoFileClip(downloaded_video_filename)
    
#     if main_video.duration >= 60:
#         print('Sorry, main video is 60 seconds or longer...')
#         return

#     # 1. Rescale overlay to main_video width
#     overlay_video = VideoFileClip(overlay_video_filename)
#     random_start = random.randint(0, (int) (overlay_video.duration - main_video.duration))
#     overlay_video = overlay_video.subclip(random_start, random_start + main_video.duration)
#     overlay_video.write_videofile(OVERLAY_CROPPED_VIDEO)

#     resize(OVERLAY_CROPPED_VIDEO, OVERLAY_RESCALED_VIDEO, new_width = main_video.w)
#     overlay_rescaled_video = VideoFileClip(OVERLAY_RESCALED_VIDEO)

#     # 2. Join videos
#     final_clip = CompositeVideoClip([main_video.set_position((0, -100)), overlay_rescaled_video.set_position('bottom')])
#     final_clip.write_videofile(output_video_filename, fps = 30)

#     # 3. Store in database that video as created
#     # db_insert('tiktok_videos', ['video_id', 'username', 'status'], [tiktok_video_data['video_id'], tiktok_video_data['username'], 'generated'])

#     #  TODO: Remove files
#     try:
#         os.remove(OVERLAY_CROPPED_VIDEO)
#         os.remove(OVERLAY_RESCALED_VIDEO)
#         #os.remove(downloaded_video_filename)
#     except:
#         print('Error deleting files')

# def resize(video_filename, output_filename = 'video_resized.mp4', new_width = None, new_height = None):
#     if not new_width and not new_height:
#         return
    
#     clip = VideoFileClip(video_filename)

#     if new_width and not new_height:
#         new_height = math.ceil((clip.h * new_width) / clip.w)
#     elif new_height and not new_width:
#         new_width = math.ceil((clip.w * new_height) / clip.h)

#     # moviepy resize method does not allow odd numbers
#     if new_width % 2 != 0:
#         new_width -= 1
#     if new_height % 2 != 0:
#         new_height -= 1
    
#     clip.resize(width = new_width, height = new_height).write_videofile(output_filename)

# def deeply_process_green_screen(original_filename, output_filename):
#     """
#     I receive a green screen transition 'original_filename' and I process it to turn
#     that green screen into a black background to, later, improve processing it with
#     mask filter.

#     This method is useful for clean transitions (maybe with not many colors). This
#     need more testing.
#     """
#     TMP_VIDEO_FILE = create_tmp_filename('tmp_deep_transition_without_audio.avi')
#     TMP_AUDIO_FILE = create_tmp_filename('tmp_deep_transition_audio.mp3')

#     # TODO: Sorry but this time output must be .avi, thank you
#     if not output_filename.endswith('.avi'):
#         print('Sorry, output must be .avi')
#         exit()

#     cap = cv2.VideoCapture(original_filename)
#     # grab one frame
#     _, frame = cap.read()
#     h,w = frame.shape[:2]
#     h = int(h)
#     w = int(w)
    
#     res = (w, h)
#     # Videowriter to build the output
#     fourcc = cv2.VideoWriter_fourcc(*'XVID')
#     out = cv2.VideoWriter(TMP_VIDEO_FILE, fourcc, 30.0, res)
    
#     # We only git the most common color in first frame (it should be the green)
#     green_screen_color = None
#     done = False
#     while not done:
#         # Get the current frame
#         ret, img = cap.read()
#         if not ret:
#             done = True
#             continue

#         # Turn into HSV
#         hsv = cv2.cvtColor(img, cv2.COLOR_BGR2HSV)
#         h,s,v = cv2.split(hsv)

#         # I choose manually (only in the first frame) the 
#         biggest = -1
#         if green_screen_color == None:
#             # Get unique colors in frame
#             unique_colors, counts = np.unique(s, return_counts=True)

#             for a in range(len(unique_colors)):
#                 if counts[a] > biggest:
#                     biggest = counts[a]
#                     green_screen_color = int(unique_colors[a])

#         # This margin was manually chosen (according to results)
#         margin = 70
#         mask = cv2.inRange(s, green_screen_color - margin, green_screen_color + margin)

#         # Smooth out the mask and invert
#         kernel = np.ones((3,3), np.uint8)
#         mask = cv2.dilate(mask, kernel, iterations = 1)
#         mask = cv2.medianBlur(mask, 5)
#         mask = cv2.bitwise_not(mask)

#         # Crop out the image and fill with color (black) to make it work well
#         #crop = np.full_like(img, [0, 255, 0])
#         crop = np.zeros_like(img)
#         # I think here is filling with BLACK, maybe I can fill it with pure GREEN
#         crop[mask == 255] = img[mask == 255]
#         out.write(crop)

#     # Free resources
#     cap.release()
#     out.release()

#     # Append audio we lost
#     extract_audio_from_video(original_filename, TMP_AUDIO_FILE)
#     append_audio(TMP_VIDEO_FILE, TMP_AUDIO_FILE, output_filename)

#     # Here we have the file stored with the green background now as black background
#     # TODO: Maybe we can use that mask to get a transparency layer with another library
#     # as opencv does not support it (as I read)
#     return True

# def append_audio(video_filename, audio_filename, output_filename = 'with_audio.mp4'):
#     """
#     Receives a video, replaces its audio with the audio in 'input_audio' file
#     and creates a new video with the 'output_video' name.
#     """
#     # TODO: Do this with an array of accepted extensions, please
#     if not video_filename.endswith('.mp4') and not video_filename.endswith('.mov') and not video_filename.endswith('.avi') and not video_filename.endswith('webm'):
#         video_filename += '.mp4'
    
#     # TODO: Create an 'accepted_video_extensions' and 'accepted_audio_extensions' and use them
#     if not audio_filename.endswith('.mp3') and not audio_filename.endswith('.wav') and not audio_filename.endswith('.m4a'):
#         audio_filename += '.mp3'

#     if output_filename == video_filename:
#         print('Output video name cannot be the same as input!')
#         return None
    
#     # TODO: Investigate a way of appending with ffmpeg because it
#     # seems to be so much faster than moviepy

#     #from subprocess import run
#     #parameters = ['ffmpeg', '-i', video_filename, '-i', audio_filename, '-c:v', 'libvpx', '-crf', '15', #'-b:v', '1M', 'copy', output_filename]
#     #run(parameters)
#     #ffmpeg -i video.mp4 -i audio.m4a -c copy output.mp4


#     videoclip = VideoFileClip(video_filename)
#     # This is a fix of bad audio ending due to ffmpeg error with moviepy
#     # See this: https://github.com/Zulko/moviepy/issues/1936
#     #audiounfixedclip = AudioFileClip(input_audio)
#     #audiounfixedclip = audiounfixedclip.subclip(0, audiounfixedclip.duration)
#     #audioclip = CompositeAudioClip([audiounfixedclip])
#     audioclip = CompositeAudioClip([AudioFileClip(audio_filename)])
#     videoclip.audio = CompositeAudioClip([audioclip])

#     videoclip.to_videofile(
#         output_filename,
#         #codec = "libx264",
#         #temp_audiofile = WIP_FOLDER + 'temp-audio.m4a',
#         #remove_temp = True,
#         #audio_codec = 'aac' # pcm_s16le or pcm_s32le 'aac' previously
#     )


from yta_general_utils.web.scraper.chrome_scraper import google_translate
from yta_general_utils.constants import Language
from yta_general_utils.programming.validator.parameter import ParameterValidator


class GoogleTranslator:
    """
    Class to wrap the functionality that allows
    translating a text by using the Google Translator
    official web page with a web scrapper.
    """

    @staticmethod
    def translate(
        text: str,
        input: Language = Language.ENGLISH,
        output: Language = Language.SPANISH
    ) -> str:
        """
        Get the translation of the given 'text' from the
        provided 'input' language to the 'output' language.
        """
        ParameterValidator.validate_mandatory_string('text', text, do_accept_empty = False)
        input = Language.to_enum(input)
        output = Language.to_enum(output)

        return google_translate(text, input.value, output.value)


"""
In this file we handle Facebook url and we parse
them to obtain basic information and check if
they are valid ones or not.

But by now I'm only accepting reel urls
- Short (shared reel): 'https://www.instagram.com/reel/DHQf6RmMFtf/?igsh=ZBDzeTA4cWkwbW4w'
"""
from yta_general_utils.programming.regular_expressions import GeneralRegularExpression
from yta_general_utils.programming.validator.parameter import ParameterValidator
from yta_general_utils.web.instagram.dataclasses import InstagramUrl


class InstagramUrlParser:
    """
    Class to simplify the way we parse Instagram
    videos urls.
    """

    @staticmethod
    def is_valid(
        url: str
    ) -> bool:
        """
        Check if the provided Instagram video 'url' is
        valid or not. A valid Instagram url can be a 
        short url or a long url.
        """
        ParameterValidator.validate_mandatory_string('url', url, do_accept_empty = False)
        
        return _is_short_instagram_url(url)
    
    @staticmethod
    def parse(
        url: str
    ) -> InstagramUrl:
        """
        Parse the provided 'url' and return a InstagramUrl
        dataclass instance containing the author username,
        the video id and the long-format url, or raises
        an Exception if the given 'url' is not valid.
        """
        ParameterValidator.validate_mandatory_string('url', url, do_accept_empty = False)

        if not InstagramUrlParser.is_valid(url):
            raise Exception('The provided "url" is not a valid Instagram video url.')
        
        url = _clean(url)
        aux = url.split('/')

        return InstagramUrl(
            video_id = aux[len(aux) - 1],
            url = url
        )
    
def _is_short_instagram_url(
    url: str
) -> bool:
    ParameterValidator.validate_mandatory_string('url', url, do_accept_empty = False)
    
    return GeneralRegularExpression.INSTAGRAM_SHORT_REEL_URL.parse(url)

def _clean(
    url: str
) -> bool:
    """
    Removes any additional parameter that is after a
    question mark sign.
    """
    ParameterValidator.validate_mandatory_string('url', url, do_accept_empty = False)
    
    return (
        url.split('?')[0]
        if '?' in url else
        url
    )


from dataclasses import dataclass


@dataclass
class FacebookUrl:

    video_id: str
    """
    The id of the video.
    """
    url: str
    """
    The short Facebook url of that video.
    """

    def __init__(
        self,
        video_id: str,
        url: str
    ):
        self.video_id = video_id
        self.url = url


from yta_general_utils.file.checker import FileValidator
from yta_general_utils.programming.validator import PythonValidator
from yta_general_utils.file.enums import FileTypeX
from moviepy import VideoFileClip
from pydub import AudioSegment
from PIL import Image
from typing import Union

import json
import io


class FileReader:
    """
    Class to simplify and encapsulate the functionality related
    with reading files.
    """

    @staticmethod
    def read_json(
        filename: str
    ):
        """
        Reads the provided 'filename' and returns the information 
        as a json (if possible).

        Parameters
        ----------
        filename : str
            File path from which we want to read the information.
        """
        if not PythonValidator.is_string(filename) or not FileValidator.file_exists(filename):
            raise Exception('The provided "filename" is not a valid string or filename.')
        
        with open(filename, encoding = 'utf-8') as json_file:
            return json.load(json_file)
        
    @staticmethod
    def read_lines(
        filename: str
    ):
        """
        Read the content of the provided 'filename'
        if valid and return it as it decomposed in
        lines.

        Parameters
        ----------
        filename : str
            File path from which we want to read the information.
        """
        if not PythonValidator.is_string(filename) or not FileValidator.file_exists(filename):
            raise Exception('The provided "filename" is not a valid string or filename.')
        
        with open(filename, 'r', encoding = 'utf-8') as file:
            return file.readlines()
        
    @staticmethod
    def read(
        filename: str
    ):
        """
        Read the content of the provided 'filename'
        if valid and return it as it is.

        Parameters
        ----------
        filename : str
            File path from which we want to read the information.
        """
        if not PythonValidator.is_string(filename) or not FileValidator.file_exists(filename):
            raise Exception('The provided "filename" is not a valid string or filename.')
        
        with open(filename, 'r', encoding = 'utf-8') as file:
            return file.read()

    @staticmethod
    def parse_file_content(
        file_content: Union[bytes, bytearray, io.BytesIO],
        file_type: FileTypeX
    ) -> Union[VideoFileClip, str, AudioSegment, Image.Image]:
        """
        Parse the provided 'file_content' with the given
        'file_type' and return that content able to be
        handled.

        This method is capable to detect videos, subtitles,
        audio, text and images.
        """
        if not PythonValidator.is_instance(file_content, [bytes, bytearray, io.BytesIO]):
            raise Exception('The provided "file_content" parameter is not bytes or bytearray.')
        
        file_type = FileTypeX.to_enum(file_type)
        
        if PythonValidator.is_instance(file_content, [bytes, bytearray]):
            # If bytes, load as a file in memory
            file_content = io.BytesIO(file_content)

        parse_fn = {
            FileTypeX.VIDEO: lambda file_content: VideoFileClip(file_content),
            FileTypeX.SUBTITLE: lambda file_content: file_content.getvalue().decode('utf-8'),
            FileTypeX.TEXT: lambda file_content: file_content.getvalue().decode('utf-8'),
            FileTypeX.AUDIO: lambda file_content: AudioSegment.from_file(file_content),
            FileTypeX.IMAGE: lambda file_content: Image.open(file_content)
        }.get(file_type, None)

        return parse_fn(file_content) if parse_fn else None

    @staticmethod
    def parse_filename(
        filename: str,
    ) -> Union[VideoFileClip, str, AudioSegment, Image.Image]:
        """
        Identify the provided 'filename' extension and open
        it according to the detected file type.

        This method is capable to detect videos, subtitles,
        audio, text and images.
        """
        reader_fn = {
            FileTypeX.VIDEO: lambda filename: VideoFileClip(filename),
            FileTypeX.SUBTITLE: lambda filename: FileReader.read(filename),
            FileTypeX.TEXT: lambda filename: FileReader.read(filename),
            FileTypeX.AUDIO: lambda filename: AudioSegment.from_file(filename),
            FileTypeX.IMAGE: lambda filename: Image.open(filename)
        }.get(FileTypeX.get_type_from_filename(filename), None)

        return reader_fn(filename) if reader_fn else None


"""
THIS IS VERY IMPORTANT TO UNDERSTAND THE CODE:

- First of all, the distance. Imagine the graphic
once it's been plotted. If you go from the origin
to the end, there is a distance you are virtually
drawing. That is the global distance. The distance
from the origin, normalized, is d=1.0. The origin
is d=0.0 and the end is d=1.0.

- As you are using pairs of nodes, there is also a
local distance in between each pair of nodes.
Imagine, for a second, that each pair of nodes is
a graphic by itself. The first node is the d=0.0
and the second node is the d=1.0 (in local distance
terms).

- If you have, for example, 5 pairs of nodes, as
the total global distance is d=1.0, each pair of
nodes will represent a 1/5 of that total global
distance, so each pair of nodes local distance
is a 1/5 = 0.2 of the total global distance.
Knowing that, in order, the pairs of nodes
represent the global distance as follows:
[0.0, 0.2], (0.2, 0.4], (0.4, 0.6], (0.6, 0.8]
and (0.8, 1.0] (for the same example with 5 pairs
of nodes).

- Now that you have clear the previous steps, if
you think about a global distance of d=0.3, that
will be in the second pair of nodes (the one that
represents (0.2, 0.4] range. But that d=0.3 is in
terms of global distance, so we need to adapt it
to that pair of nodes local distance. As we
skipped 1 (the first) pair of nodes, we need to
substract its distance representation from our
global distance, so d=0.3 - 0.2 => d=0.1. Now, as
d=0.1 is a global distance value, we need to turn
it into a local distance value. As each pair of
nodes represents a 0.2 of the global distance, we
do this: d=0.1 / 0.2 => d=0.5 and we obtain a
local distance of d=0.5. That is the local
distance within the second pair of nodes (in this
example) we need to look for.

- Once we know the local distance we need to 
look for, as we now the pair of nodes X value of
each of those nodes, we can calculate the
corresponding X value for what that local distance
fits.

As you can see, we go from a global distance to
a local X value to obtain the corresponding Y
of the affected pair of nodes. This class has
been created with the purpose of following
an animation progress, so the distance we are
talking about is actually the amount of that
animation we have done previously so we can obtain
the next value we need to apply to obtain the
animation the graphic describes.
"""
from yta_general_utils.math.graphic.graphic_axis import GraphicAxis
from yta_general_utils.math.graphic.node import Node
from yta_general_utils.math.graphic.pair_of_nodes import PairOfNodes
from yta_general_utils.programming.validator import PythonValidator
from yta_general_utils.programming.validator.number import NumberValidator
from yta_general_utils.math.progression import Progression

import matplotlib.pyplot as plt


class Graphic:
    """
    Class that represent a Graphic in which we will
    place Nodes and apply rate functions between
    those nodes to calculate the corresponding y
    values.

    This class allows Nodes that are not placed in
    the same position. The Nodes will be ordered
    automatically by their 'x' position, but you can
    add them unordered (but unrepeated).

    This Graphic class has been created to be able
    to use them as a custom parameter modifier as
    we are able to build the Graphic we want and
    apply it to any parameter.
    """
    
    x_axis: GraphicAxis = None
    """
    The x axis which contains the min and max x 
    valid values.
    """
    y_axis: GraphicAxis = None
    """
    The y axis which contains the min and max y
    valid values.
    """
    _nodes: list[Node] = None
    """
    The list of nodes defined in the graphic to
    build it. These nodes are interconnected with
    a rate function.
    """

    @property
    def nodes(self):
        """
        Get the nodes ordered by the x position.
        """
        return sorted(self._nodes, key = lambda node: node.position[0])
    
    @property
    def min_x(self):
        return min(self.nodes, key = lambda node: node.x).x
    
    @property
    def max_x(self):
        return max(self.nodes, key = lambda node: node.x).x
    
    @property
    def min_y(self):
        return min(self.nodes, key = lambda node: node.y).y
    
    @property
    def max_y(self):
        return max(self.nodes, key = lambda node: node.y).y
    
    @property
    def pairs_of_nodes(self) -> list[PairOfNodes]:
        """
        Get pairs of nodes ordered by the x position. There
        is, at least, one pair of nodes that will be, if no
        more nodes added, the first and the last one.
        """
        return [
            PairOfNodes(self.nodes[index], self.nodes[index + 1])
            for index, _ in enumerate(self.nodes[1:])
        ]

    def __init__(self, x_axis: GraphicAxis, y_axis: GraphicAxis):
        if not PythonValidator.is_instance(x_axis, GraphicAxis) or not PythonValidator.is_instance(y_axis, GraphicAxis):
            raise Exception('The "x_axis" and "y_axis" parameter must be instances of GraphicAxis class.')

        # TODO: Maybe make it possible to be instantiated with a
        # list of nodes
        
        self.x_axis = x_axis
        self.y_axis = y_axis
        self._nodes = []

    def add_node(self, x: float, y: float):
        """
        Add a new node to the graphic if its position is
        inside the x and y axis ranges and if there is not
        another node in that position.

        This method returns the instance so you can chain
        more than one 'add_node' method call.
        """
        if x < self.x_axis.min or x > self.x_axis.max:
            raise Exception(f'The provided "x" is not in the [{self.x_axis.min}, {self.x_axis.max}] range.')
        
        if y < self.y_axis.min or y > self.y_axis.max:
            raise Exception(f'The provided "y" is not in the [{self.y_axis.min}, {self.y_axis.max}] range.')
        
        if self._get_node(x) is not None:
            raise Exception('There is another node in the provided "x" position.')

        self._nodes.append(Node(x, y))

        return self

    def get_n_values(
        self,
        n: int
    ):
        """
        Return a list of 'n' (x, y) values of the graphic
        from the start to the end of it.
        """
        return [
            self.get_xy_from_normalized_d(d)
            for d in Progression(0, 1, n).values
        ]

    def get_xy_from_normalized_d(self, d: float):
        """
        Return a not normalized tuple (x, y) representing
        the position for the provided global distance 'd',
        that is the global distance in the whole graphic X
        axis. The lower (left) 'x' of the graphic would be
        d=0.0 and the highest (right) 'x', d=1.0.
        """
        x = self._get_x_from_normalized_d(d)
        _, y = self.get_xy_from_not_normalized_x(x)

        return (x, y)
    
    def get_xy_from_not_normalized_x(self, x: float) -> tuple[float, float]:
        """
        Return a not normalized tuple (x, y) representing
        the position for the provided 'x' of the graphic.

        The returned 'x' value will be the same as the one
        provided as parameter.
        """
        if not NumberValidator.is_number_between(x, self.min_x, self.max_x):
            raise Exception(f'The "x" parameter is out of the graphic bounds, it must be between [{self.min_x}, {self.max_x}].')
        
        return (x, self._get_pair_of_node_from_x(x).get_y_from_not_normalized_x(x).value)
    
    def _get_node(self, x: float):
        """
        Get the node placed at the provided 'x' if existing
        or None if not. This method is used internally to
        verify that there is not another node at the same
        position when adding one.
        """
        return next((node for node in self.nodes if node.x == x), None)
    
    def _get_pair_of_node_from_x(self, x: float):
        """
        Obtain the pair of nodes in which the provided 'x'
        is contained.
        """
        return next((pair_of_node for pair_of_node in self.pairs_of_nodes if x <= pair_of_node.max_x), None)

    def _get_x_from_normalized_d(self, d: float):
        """
        Obtain the not normalized 'x' value for the
        provided 'd', that is the global distance in the
        whole graphic X axis. The lower (left) 'x' of
        the graphic would be d=0.0 and the highest
        (right) 'x', d=1.0.

        In the graphic, the distance from the first 'x'
        to the last 'x' is d=1.0. The first (lower) 'x'
        value is d=0.0 and the last (highest) 'x' value
        is d=1.0. The graphic contains more than one 
        pair of nodes, so the actual X to which this 
        global 'd' refers to is in one of those pairs
        of nodes.

        If a graphic contains 5 pairs of nodes, as the
        global distance is d=1.0, each pair of nodes
        represents a d=0.2. First pair of nodes from
        [0.0, 0.2], second pair (0.2, 0.4], etc. So,
        if a global distance d=0.3 is requested, the
        second pair of nodes (in the example below) 
        will be used to calculate its X and the 
        corresponding Y value.
        """
        if not NumberValidator.is_number_between(d, 0.0, 1.0):
            raise Exception(f'The provided "d" parameter "{str(d)}" must be between 0.0 and 1.0, both inclusive.')
        
        # 'd' is the distance within the whole graphic
        # representation with a value between [0.0, 1.0]
        num_of_pairs_of_nodes = len(self.pairs_of_nodes)
        # This is the distance that each pair of nodes is
        # representing (also between [0.0, 1.0]). As a
        # reminder, if 5 pairs of nodes, 1.0 / 5 = 0.2, so
        # each pair of nodes would be representing a 0.2
        # portion of the whole graphic distance.
        pair_of_nodes_d = 1 / num_of_pairs_of_nodes
        # Same example as above, 5 pairs of nodes:
        # d = 0.2 // 0.2 = index 1 which is for the 2nd pair
        # d = 0.3 // 0.2 = index 1 which is for the 2nd pair
        # d = 0.7 // 0.2 = index 3 which is for the 4th pair
        # TODO: Maybe let the last value (exact, so % = 0) for
        # the previous pair instead of for the next one (?)
        pair_of_nodes_index = int(d // pair_of_nodes_d) if d != 1.0 else num_of_pairs_of_nodes - 1
        pair_of_nodes = self.pairs_of_nodes[pair_of_nodes_index]

        # If d=0.3, we will use the 2nd pair of nodes, but that
        # as we skipped the first pair of points, we need to
        # substract the corresponding amount, so d=0.3 - 0.2 =>
        # d=0.1. Now, working locally in the 2nd pair of nodes,
        # we are looking for the d=0.1, which is a general
        # distance that we need to turn into a local one. As
        # the pair of nodes is representing a total of 0.2d of
        # the total graphic distance, that d=0.1 represents 
        # the 0.5 in the local distance of that pair of nodes:
        # d=0.2 = d=1.0 locally => 0.1 * 1 / 0.2 = 0.1 / 0.2 = 0.5
        # so the 10% (0.1d) in global terms means a 50% (0.5d) in
        # local pair of nodes terms.
        # Formula explained with the same example below:
        # X = 0.3 % (1 * 0.2) / 0.2
        # X = 0.3 % 0.2 / 0.2
        # X = 0.1 / 0.2
        # X = 0.5
        if d != 1.0:
            d = (
                d % (pair_of_nodes_index * pair_of_nodes_d) / pair_of_nodes_d
                if pair_of_nodes_index > 0 else d / pair_of_nodes_d
            )
        # Now we need to ask the pair of nodes to calculate the Y
        # position of the X that is in the 'd' local distance:
        # y = min_x + d * (max_x - min_x)
        x = pair_of_nodes.left_node.x + d * (pair_of_nodes.right_node.x - pair_of_nodes.left_node.x)

        return x
    
    def plot(self):
        # Limit and draw axis
        plt.xlim(self.x_axis.min, self.x_axis.max)
        plt.ylim(self.y_axis.min, self.y_axis.max)
        plt.axhline(0, color = 'black', linewidth = 1)
        plt.axvline(0, color = 'black', linewidth = 1)

        plt.grid(True)

        # Draw nodes
        x_vals = [node.x for node in self.nodes]
        y_vals = [node.y for node in self.nodes]
        plt.scatter(x_vals, y_vals, color = 'white', edgecolors = 'black', s = 100)

        # Draw points between nodes
        xs = []
        ys = []
        for pair_of_node in self.pairs_of_nodes:
            positions = pair_of_node.get_n_xy_values_to_plot(100)
            t_xs, t_ys = zip(*positions)
            xs += t_xs
            ys += t_ys
       
        plt.scatter(xs, ys, color = 'black', s = 1)
        
        plt.title('')
        plt.xlabel('X axis')
        plt.ylabel('Y axis')
        
        plt.show()


from yta_general_utils.programming.dataclasses import MethodParameters, MethodParameter
from yta_general_utils.programming.validator.parameter import ParameterValidator
from typing import Union

import inspect


class ParameterObtainer:
    """
    Class to interact with python methods and classes to obtain
    the parameters those method have. This is usefull for dynamic
    functionality that need to fill or check if the required
    parameters are passed or not.
    """

    PARAMETER_EMPTY = inspect._empty
    """
    Class that represents an empty parameter default value
    """
    
    @staticmethod
    def get_parameters_from_method(
        method: callable,
        parameters_to_ignore: Union[list[str], None] = []
    ) -> MethodParameters:
        """
        Obtain the parameters of the given 'method' as
        a list of our custom dataclass Parameter, easy
        to handle and to get information from.
        """
        ParameterValidator.validate_mandatory_callable('method', method)
        ParameterValidator.validate_list_of_string('parameters_to_ignore', parameters_to_ignore)
        
        # If arrays are None by any chance, turn into empty arrays
        parameters_to_ignore = [] if parameters_to_ignore is None else parameters_to_ignore

        # We will force these normal parameters to ignore them
        parameters_to_ignore += ['self', 'cls', 'args', 'kwargs']

        return MethodParameters([
            MethodParameter(
                method_parameter.name,
                method_parameter.annotation,
                method_parameter.default
            )
            for method_parameter in inspect.signature(method).parameters.values()
            if method_parameter.name not in parameters_to_ignore
        ])

"""
This code below is nice for testing:
from typing import Union
import inspect

def test_method(text: str, output_filename: Union[str, int, None] = None):
    pass

for parameter in inspect.signature(test_method).parameters:
    print(parameter)

for parameter_value in inspect.signature(test_method).parameters.values():
    print('Este es uno:')
    print(parameter_value.name)
    #print(parameter_value.kind)
    print(parameter_value.annotation)
    print(parameter_value.default)
"""


"""
Class to dynamically obtain the attributes in
an instance.
"""

class AttributeObtainer:
    """
    Class to interact with an instance and dynamically obtain
    the attributes that have been set. This is useful to 
    dynamically call methods that need those parameters.
    """

    @staticmethod
    def get_attributes_from_instance(
        instance: object,
        attributes_to_ignore: list[str] = []
    ):
        """
        Obtain the attributes of the given 'instance'
        as a dict containing the keys and values of
        these attributes.

        Only the values that are actually set on the
        instance are obtained with 'vars'. If you set
        'var_name = None' but you don't do 
        'self.var_name = 33' in the '__init__' method,
        it won't be returned by the 'vars()' method.
        """
        # Each attribute is a dict 'key: value'
        return {
            k: v
            for k, v in vars(instance).items()
            if k not in attributes_to_ignore
        }



"""
In this file we handle Facebook url and we parse
them to obtain basic information and check if
they are valid ones or not.

We have different types of valid urls:
- 'https://www.facebook.com/watch/?v=499831560618548'
- 'https://www.facebook.com/RolandGarros/videos/10155404760334920/FOO'
- 'https://www.facebook.com/RolandGarros/videos/FOO/10155404760334920'

But by now I'm only accepting reel urls
- Short (shared reel): 'https://www.facebook.com/share/r/1ZBKSvZZVr/'
"""
from yta_general_utils.programming.regular_expressions import GeneralRegularExpression
from yta_general_utils.programming.validator.parameter import ParameterValidator
from yta_general_utils.web.facebook.dataclasses import FacebookUrl


class FacebookUrlParser:
    """
    Class to simplify the way we parse Facebook
    videos urls.
    """

    @staticmethod
    def is_valid(
        url: str
    ) -> bool:
        """
        Check if the provided Facebook video 'url' is
        valid or not. A valid Facebook url can be a 
        short url or a long url.
        """
        ParameterValidator.validate_mandatory_string('url', url, do_accept_empty = False)
        
        return _is_short_facebook_url(url)
    
    @staticmethod
    def parse(
        url: str
    ) -> FacebookUrl:
        """
        Parse the provided 'url' and return a FacebookUrl
        dataclass instance containing the video id and the
        short-format url, or raises an Exception if the
        given 'url' is not valid.
        """
        ParameterValidator.validate_mandatory_string('url', url, do_accept_empty = False)

        if not FacebookUrlParser.is_valid(url):
            raise Exception('The provided "url" is not a valid Facebook video url.')
        
        url = _clean(url)
        aux = url.split('/')

        return FacebookUrl(
            video_id = aux[len(aux) - 1],
            url = url
        )
    
def _is_short_facebook_url(
    url: str
) -> bool:
    ParameterValidator.validate_mandatory_string('url', url, do_accept_empty = False)
    
    return GeneralRegularExpression.FACEBOOK_SHORT_REEL_URL.parse(url)

def _clean(
    url: str
) -> str:
    """
    Removes any additional parameter that is after a
    question mark sign.
    """
    ParameterValidator.validate_mandatory_string('url', url, do_accept_empty = False)
    
    return (
        url.split('?')[0]
        if '?' in url else
        url
    )


from yta_general_utils.file.filename import get_file_extension
from yta_general_utils.file.writer import FileWriter
from yta_general_utils.programming.output import Output
from lxml.html import fromstring
from typing import Union

import requests


# TODO: Move to a 'consts.py' (?)
DRIVE_RESOURCE_URL_START = 'https://drive.google.com/file/d/'
DOWNLOAD_URL = 'https://docs.google.com/uc?export=download&confirm=1'

class GoogleDriveResource:
    """
    Class to handle Google Drive Resources. Just instantiate it with its
    Google Drive url and it will be ready for download if the url is valid
    and available.

    A valid 'drive_url' must be like this:
    https://drive.google.com/file/d/1rcowE61X8c832ynh0xOt60TH1rJfcJ6z/view?usp=sharing&confirm=1
    """
    url: str = ''
    filename: str = ''
    id: str = ''

    def __init__(self, drive_url: str):
        """
        Initializes this instance and sets the 'url', the real 'filename'
        obtained from Google Drive and also the 'id' obtained from the url.
        """
        self.url = self.parse_url(drive_url)
        self.filename = self.get_filename(self.url)
        self.id = self.get_id(self.url)

    def download(
        self,
        output_filename: str
    ):
        """
        Downloads the Google Drive Resource to the provided 'output_filename'.
        If not 'output_filename' parameter provided, the system will generate
        a temporary file. The downloaded file filename will be returned.
        """
        return self._download_resource(self.id, self.filename, output_filename)

    @classmethod
    def parse_url(cls, drive_url: str):
        """
        This method validates if the provided 'drive_url' is a valid
        Google Drive url and returns it if valid or raises an Exception
        if invalid.
        """
        drive_url = drive_url.replace('http://', 'https://')

        if not drive_url.startswith('https://'):
            drive_url = f'https://{drive_url}'

        if not drive_url.startswith(DRIVE_RESOURCE_URL_START):
            raise Exception(f'Provided "google_drive_url" parameter {drive_url} is not valid. It must be like "{DRIVE_RESOURCE_URL_START}..."')
    
        if not 'confirm=t' in drive_url:
            # previously was '&confirm=1' to avoid virus scan as they say:
            # https://github.com/tensorflow/datasets/issues/3935#issuecomment-2067094366
            drive_url += '&confirm=t'

        return drive_url

    @classmethod
    def is_url_available(cls, drive_url: str):
        """
        This method validates if the provided 'drive_url' is available
        for access and download.
        """
        # TODO: Check somehow that url is available or raise Exception if not
        return True
        drive_file_name = fromstring(requests.get(drive_url).content).findtext('.//title').split('-')[0].strip()

    @classmethod
    def get_id(cls, drive_url: str):
        """
        Parses the 'drive_url' and returns the Google Drive Resource id
        if url is valid.
        """
        drive_url = cls.parse_url(drive_url)

        return drive_url.replace(DRIVE_RESOURCE_URL_START, '').split('/')[0]

    @classmethod
    def get_filename(cls, drive_url: str):
        """
        Parses the 'drive_url' and returns the Google Drive Resource 
        filename if url is valid. This is the real filename in Google
        Drive.
        """
        drive_url = cls.parse_url(drive_url)

        return fromstring(requests.get(drive_url).content).findtext('.//title').split('-')[0].strip()

    @classmethod
    def download_from_url(cls, drive_url: str, output_filename: str = None):
        drive_url = cls.parse_url(drive_url)
        id = cls.get_id(drive_url)
        filename = cls.get_filename(drive_url)

        return cls._download_resource(id, filename, output_filename)
    
    @classmethod
    def _download_resource(
        cls,
        id: str,
        filename: str,
        output_filename: Union[str, None] = None
    ):
        session = requests.Session()
        # Trying to obtain the web title to get the file name
        response = session.get(DOWNLOAD_URL, params = {'id': id}, stream = True)

        # Look for a token to be able to download
        token = None
        for key, value in response.cookies.items():
            if key.startswith('download_warning'):
                token = value
                break

        if token:
            params = {'id': id, 'confirm': token}
            response = session.get(DOWNLOAD_URL, params = params, stream = True)
            # TODO: Handle virus unchecked Google warning that contains this below:
            # <title>Google Drive - Virus scan warning</title>
            # check Notion for the entire error message

        output_filename = Output.get_filename(
            output_filename,
            get_file_extension(filename)
        )
        
        # Save response
        FileWriter.write_file_by_chunks_from_response(response, output_filename)

        return output_filename
    
        
# TODO: Remove these methods below when GoogleDriveResource working
def download_file_from_google_drive(drive_url: str, output_filename: str = None):
    """
    @deprecated 
    Downloads a file from a given Google Drive url that must be
    generated by sharing the file and making it public (this
    means anyone can view and download it).

    The file will be stored locally as 'output_filename', but 
    the extension could be changed if the provided one is
    different than the one that the Google Drive file has.
    This final created file name will be returned.

    A valid 'drive_url' must be like this:
    https://drive.google.com/file/d/1rcowE61X8c832ynh0xOt60TH1rJfcJ6z/view?usp=sharing&confirm=1
    """
    # TODO: Check if drive_url is valid and downloadable (regex?)

    # This is to avoid virus scan
    if not 'confirm=t' in drive_url:
        # previously was '&confirm=1' to avoid virus scan as they say:
        # https://github.com/tensorflow/datasets/issues/3935#issuecomment-2067094366
        drive_url += '&confirm=t'

    # Lets get file id and file name
    drive_file_id = drive_url.replace('https://drive.google.com/file/d/', '').split('/')[0]
    # We fire a simple get request to obtain file name from url title
    drive_file_name = fromstring(requests.get(drive_url).content).findtext('.//title').split('-')[0].strip()

    DOWNLOAD_URL = "https://docs.google.com/uc?export=download&confirm=1"

     # Lets download the file
    session = requests.Session()
    
    # Trying to obtain the web title to get the file name
    response = session.get(DOWNLOAD_URL, params = {'id': drive_file_id}, stream = True)

    # Look for a token to be able to download
    token = None
    for key, value in response.cookies.items():
        if key.startswith('download_warning'):
            token = value
            break

    if token:
        params = {'id': drive_file_id, 'confirm': token}
        response = session.get(DOWNLOAD_URL, params = params, stream = True)
        # TODO: Handle virus unchecked Google warning that contains this below:
        # <title>Google Drive - Virus scan warning</title>
        # check Notion for the entire error message

    output_filename = Output.get_filename(
        output_filename,
        get_file_extension(drive_file_name)
    )

    # Save response
    FileWriter.write_file_by_chunks_from_response(response, output_filename)

    return output_filename

# TODO: Move this to another file
def get_id_from_google_drive_url(google_drive_url: str):
    """
    @deprecated
    This method obtains the Google Drive element id and returns it if
    the provided 'google_drive_url' is valid.
    """
    if not google_drive_url:
        return None
    
    # url should be like: https://drive.google.com/file/d/1My5V8gKcXDQGGKB9ARAsuP9MIJm0DQOK/view?usp=sharing&confirm=1
    if not google_drive_url.startswith('https://drive.google.com/file/d/'):
        raise Exception(f'Provided "google_drive_url" parameter {google_drive_url} is not valid. It must be like "https://drive.google.com/file/d/..."')

    google_drive_url = google_drive_url.replace('https://drive.google.com/file/d/', '')

    if '/' in google_drive_url:
        google_drive_url = google_drive_url.split('/')[0]

    return google_drive_url
